@@ -20,6 +20,9 @@ API extensions.
 
 import os
 import base64
+import re
+import json
+import sys
 
 from novaclient import utils
 from novaclient import base
@@ -28,7 +31,18 @@ from novaclient.v1_1 import shell
 
 from . import agent
 
-CAPABILITIES = ('user-data', 'launch-name', 'security-groups', 'availability-zone')
+# Add new client capabilities here. Each key is a capability name and its value
+# is the list of API capabilities upon which it depends.
+
+CAPABILITIES = {'user-data': ['user-data'],
+                'launch-name': ['launch-name'],
+                'security-groups': ['security-groups'],
+                'num-instances': ['num-instances'],
+                'availability-zone': ['availability-zone'],
+                'bless-name': ['bless-name'],
+                'launch-key': ['launch-key'],
+                'import-export': ['import-export'],
+                }
 
 def __pre_parse_args__():
     pass
@@ -87,6 +101,8 @@ def _find_server(cs, server):
            help='User data file to pass to be exposed by the metadata server')
 @utils.arg('--security-groups', metavar='<security groups>', default=None, help='comma separated list of security group names.')
 @utils.arg('--availability-zone', metavar='<availability zone>', default=None, help='The availability zone for instance placement.')
+@utils.arg('--num-instances', metavar='<number>', default='1', help='Launch multiple instances at a time')
+@utils.arg('--key-name', metavar='<key name>', default=None, help='Key name of keypair that should be created earlier with the command keypair-add')
 @utils.arg('--params', action='append', default=[], metavar='<key=value>', help='Guest parameters to send to vms-agent')
 def do_launch(cs, args):
     """Launch a new instance."""
@@ -118,16 +134,19 @@ def do_launch(cs, args):
                                            user_data=user_data,
                                            guest_params=guest_params,
                                            security_groups=security_groups,
-                                           availability_zone=availability_zone)
+                                           availability_zone=availability_zone,
+                                           num_instances=int(args.num_instances),
+                                           key_name=args.key_name)
 
     for server in launch_servers:
         _print_server(cs, server)
 
 @utils.arg('server', metavar='<instance>', help="ID or name of the instance to bless")
+@utils.arg('--name', metavar='<name>', default=None, help="The name of the new blessed instance")
 def do_bless(cs, args):
     """Bless an instance."""
     server = _find_server(cs, args.server)
-    blessed_servers = cs.gridcentric.bless(server)
+    blessed_servers = cs.gridcentric.bless(server, args.name)
     for server in blessed_servers:
         _print_server(cs, server)
 
@@ -162,6 +181,50 @@ def do_list_blessed(cs, args):
     server = _find_server(cs, args.server)
     _print_list(cs.gridcentric.list_blessed(server))
 
+@utils.arg('server', metavar='<server>', help="ID or name of the blessed instance")
+@utils.arg('output', metavar='<output>', default=None, help="Name of a file to write the exported data to.")
+def do_gc_export(cs, args):
+    """Export a blessed instance"""
+    server = _find_server(cs, args.server)
+    result = server.export()
+
+    json_result = json.dumps(result, sort_keys=True, indent=4, separators=(',', ': ')) + '\n'
+
+    with file(args.output, 'w') as f:
+        f.write(json_result)
+
+    print "Instance data is being exported to image %s" %(result['export_image_id'])
+
+@utils.arg('data_filename', metavar='<data-filename>',
+                              help="A file containing the exported server data")
+@utils.arg('--override', metavar='<override>',
+                      help="Semicolon-separated list of parameters to override")
+def do_gc_import(cs, args):
+    """Import a blessed instance"""
+
+    # The override option can be something like this:
+    # export_image_id=THE-ID;security_groups=sg1,sg2;fields.display_name=foo
+
+    # Read in the contents of the server data (should be a JSON file)
+    with open(args.data_filename, 'r') as f:
+        data = json.load(f)
+
+    if args.override is not None:
+        def override(key, value, data):
+            if '.' in key:
+                parent, _, key = key.partition('.')
+                override(key, value, data[parent])
+            else:
+                if isinstance(data[key], list):
+                    value = value.split(',')
+                data[key] = value
+        for override_arg in args.override.split(';'):
+            key, value = override_arg.split('=', 1)
+            override(key, value, data)
+
+    server = cs.gridcentric.import_instance(data)
+    _print_server(cs, server)
+
 @utils.arg('--flavor',
      default=None,
      metavar='<flavor>',
@@ -288,24 +351,39 @@ def do_gc_boot(cs, args):
      default=None,
      metavar='<agent_version>',
      help="Install a specific agent version.")
+@utils.arg('--ip',
+     default=None,
+     metavar='<ip>',
+     help="Instance IP address to use (defaults to first ssh-able).")
 def do_gc_install_agent(cs, args):
     """Install the agent onto an instance."""
     server = _find_server(cs, args.server)
     server.install_agent(args.user,
                          args.key_path,
                          location=args.agent_location,
-                         version=args.agent_version)
+                         version=args.agent_version,
+                         ip=args.ip)
 
 class GcServer(servers.Server):
     """
     A server object extended to provide gridcentric capabilities
     """
 
-    def launch(self, target="0", name=None, user_data=None, guest_params={}, security_groups=None):
-        return self.manager.launch(self, target, name, user_data, guest_params, security_groups)
+    def launch(self, target="0", name=None, user_data=None, guest_params={},
+               security_groups=None, availability_zone=None, num_instances=1,
+               key_name=None):
+        return self.manager.launch(self,
+                                   target=target,
+                                   name=name,
+                                   user_data=user_data,
+                                   guest_params=guest_params,
+                                   security_groups=security_groups,
+                                   availability_zone=availability_zone,
+                                   num_instances=num_instances,
+                                   key_name=key_name)
 
-    def bless(self):
-        return self.manager.bless(self)
+    def bless(self, name=None):
+        return self.manager.bless(self, name)
 
     def discard(self):
         self.manager.discard(self)
@@ -319,27 +397,52 @@ class GcServer(servers.Server):
     def list_blessed(self):
         return self.manager.list_blessed(self)
 
-    def install_agent(self, user, key_path, location=None, version=None):
-        self.manager.install_agent(self, user, key_path, location=location, version=version)
+    def export(self):
+        return self.manager.export(self)
+
+    def install_agent(self, user, key_path, location=None,
+                        version=None, ip=None):
+        self.manager.install_agent(self, user, key_path, location=location,
+                                    version=version, ip=ip)
 
 class GcServerManager(servers.ServerManager):
     resource_class = GcServer
 
     def __init__(self, client, *args, **kwargs):
-        self.CAPABILITIES = CAPABILITIES
-
         servers.ServerManager.__init__(self, client, *args, **kwargs)
 
         # Make sure this instance is available as gridcentric.
         if not(hasattr(client, 'gridcentric')):
             setattr(client, 'gridcentric', self)
 
+    # Capabilities must be computed lazily because self.api.client isn't
+    # available in __init__
+
+    def setup_capabilities(self):
+        api_caps = self.get_info()['capabilities']
+        self.capabilities = [cap for cap in CAPABILITIES.keys() if \
+                all([api_req in api_caps for api_req in CAPABILITIES[cap]])]
+
+    def satisfies(self, requirements):
+        if not hasattr(self, 'capabilities'):
+            self.setup_capabilities()
+
+        return set(requirements) <= set(self.capabilities)
+
+    def get_info(self):
+        url = '/gcinfo'
+        res = self.api.client.get(url)[1]
+        return res
+
     def launch(self, server, target="0", name=None, user_data=None,
-               guest_params={}, security_groups=None, availability_zone=None):
+               guest_params={}, security_groups=None, availability_zone=None,
+               num_instances=1, key_name=None):
         params = {'target': target,
                   'guest': guest_params,
                   'security_groups': security_groups,
-                  'availability_zone': availability_zone}
+                  'availability_zone': availability_zone,
+                  'num_instances': num_instances,
+                  'key_name': key_name}
 
         if name != None:
             params['name'] = name
@@ -357,8 +460,9 @@ class GcServerManager(servers.ServerManager):
         header, info = self._action("gc_launch", base.getid(server), params)
         return [self.get(server['id']) for server in info]
 
-    def bless(self, server):
-        header, info = self._action("gc_bless", base.getid(server))
+    def bless(self, server, name=None):
+        params = {'name': name}
+        header, info = self._action("gc_bless", base.getid(server), params)
         return [self.get(server['id']) for server in info]
 
     def discard(self, server):
@@ -378,6 +482,16 @@ class GcServerManager(servers.ServerManager):
         header, info = self._action("gc_list_blessed", base.getid(server))
         return [self.get(server['id']) for server in info]
 
+    def export(self, server):
+        header, info = self._action("gc_export", server.id)
+        return info
+
+    def import_instance(self, data):
+        url = "/gc-import-server"
+        body = {'data': data}
+
+        return self._create(url, body, 'server')
+
     def create(self, name, image, flavor, meta=None, files=None,
                reservation_id=None, min_count=None,
                max_count=None, security_groups=None, userdata=None,
@@ -407,5 +521,7 @@ class GcServerManager(servers.ServerManager):
         return self._boot(resource_url, response_key, *boot_args,
                 **boot_kwargs)
 
-    def install_agent(self, server, user, key_path, location=None, version=None):
-        agent.install(server, user, key_path, location=location, version=version)
+    def install_agent(self, server, user, key_path, location=None,
+                        version=None, ip=None):
+        agent.install(server, user, key_path, location=location,
+                        version=version, ip=ip)
